//
// C++ Implementation: GameStateManager
//
// Description: 
//
//
// Author:  Oliver Scheuss, (C) 2007
//
// Copyright: See COPYING file that comes with this distribution
//
//
#include "GameStateManager.h"

namespace network {

GameStateManager::GameStateManager()
{
}

GameStateManager::~GameStateManager()
{
}

/**
 * This function goes through the whole list of synchronisables and 
 * saves all the synchronisables to a flat "list".
 * @return struct of type gamestate containing the size of the whole gamestate and a pointer linking to the flat list
 */
GameStateCompressed GameStateManager::getSnapshot(int id)
{
  //the size of the gamestate
  int totalsize=0;
  //the size of one specific synchronisable
  int tempsize=0;
  // get the start of the Synchronisable list
  orxonox::Iterator<Synchronisable> it;
  // struct for return value of Synchronisable::getData()
  syncData sync;
  
  GameState retval; //return value
  retval.id=id;
  // reserve a little memory and increase it later on
  retval.data = (unsigned char*)malloc(1);
  
  // offset of memory functions
  int offset=0;
  // go through all Synchronisables
  for(it = orxonox::ObjectList<Synchronisable>::start(); it != 0; ++it){
    //get size of the synchronisable
    tempsize=it->getSize();
    // add place for data and 3 ints (length,classid,objectid)
    totalsize+=tempsize+3*sizeof(int);
    // allocate additional space
    retval.data = (unsigned char *)realloc((void *)retval.data, totalsize);
    
    // run Synchronisable::getData with offset and additional place for 3 ints in between (for ids and length)
    sync=it->getData(retval.data+offset+3*sizeof(int));
    *(retval.data+offset)=sync.length;
    *(retval.data+offset+sizeof(int))=sync.objectID;
    *(retval.data+offset+2*sizeof(int))=sync.classID;
    // increase data pointer
    offset+=tempsize+3*sizeof(int);
  }
  retval.size=totalsize;
  return compress(retval);
}

/**
 * This function loads a Snapshort of the gamestate into the universe
 * @param state a GameState struct containing the size of the gamestate and a pointer linking to a flat list (returned by getSnapshot)
 */
bool GameStateManager::loadSnapshot(GameStateCompressed compstate)
{
  GameState state = decompress(compstate);
  unsigned char *data=state.data;
  // get the start of the Synchronisable list
  orxonox::Iterator<Synchronisable> it=orxonox::ObjectList<Synchronisable>::start();
  syncData sync;
  // loop as long as we have some data ;)
  while(data < state.data+state.size){
    // prepare the syncData struct
    sync.length = *(int *)data;
    data+=sizeof(int);
    sync.objectID = *(int *)data;
    data+=sizeof(int);
    sync.classID = *(int *)data;
    data+=sizeof(int);
    sync.data = data;
    data+=sync.length;
    
    if(it->objectID!=sync.objectID){
      // bad luck ;)
      // delete the synchronisable (obviously seems to be deleted on the server)
      while(it != 0 && it->objectID!=sync.objectID){
        removeObject(it);
      }
      if(it==0){  // add the new object
        // =================== factory command to add object
        // can we be sure the object really was added?
        it=orxonox::ObjectList<Synchronisable>::end();
        it->objectID=sync.objectID;
        it->classID=sync.classID;
      }
    } else {
      // we have our object
      if(! it->updateData(sync))
        std::cout << "We couldn't update objectID: " \
          << sync.objectID << "; classID: " << sync.classID << std::endl;
    }
    
  }
  
  return true;
}

/**
 * This function removes a Synchronisable out of the universe
 * @param it iterator of the list pointing to the object
 * @return iterator pointing to the next object in the list
 */
void GameStateManager::removeObject(orxonox::Iterator<Synchronisable> &it){
  orxonox::Iterator<Synchronisable> temp=it;
  ++it;
  delete  *temp;
//   return it;
}

GameStateCompressed GameStateManager::encode(GameState a, GameState b){
  GameState r = diff(a,b);
  return compress(r);
}

GameState GameStateManager::decode(GameState a, GameStateCompressed x){
  GameState t = decompress(x);
  return diff(a, t);
}

GameState GameStateManager::diff(GameState a, GameState b){
  unsigned char *ap = a.data, *bp = b.data;
  int of=0; // pointers offset
  int dest_length=0;
  if(a.size>=b.size)
    dest_length=a.size;
  else
    dest_length=b.size;
  unsigned char *dp = (unsigned char *)malloc(dest_length*sizeof(unsigned char));
  while(of<a.size && of<b.size){
    *(dp+of)=*(ap+of)^*(bp+of); // do the xor
    ++of;
  }
  if(a.size!=b.size){ // do we have to fill up ?
    unsigned char n=0;
    if(a.size<b.size){
      while(of<dest_length){
        *(dp+of)=n^*(bp+of);
        of++;
      }
    } else{
      while(of<dest_length){
        *(dp+of)=*(ap+of)^n;
        of++;
      }
    }
  }
  // should be finished now
  GameState r = {b.id, dest_length, dp};
  return r;
}

GameStateCompressed GameStateManager::compress(GameState a) {
  int size = a.size;
  uLongf buffer = (uLongf)((a.size + 12)*1.01)+1;
  unsigned char* dest = (unsigned char*)malloc( buffer );
  int retval;
  retval = compress( dest, &buffer, a.data, (uLong)size );  
  
  switch ( retval ) {
  case Z_OK: cout << "successfully compressed" << endl; break;
  case Z_MEM_ERROR: cout << "not enough memory available" << endl; break;
  case Z_BUF_ERROR: cout << "not enough memory available in the buffer" << endl; break;
  case Z_DATA_ERROR: cout << "data corrupted" << endl; break;
  }
  
  GameStateCompressed compressedGamestate = new GameStateCompressed;
  compressedGamestate.compsize = buffer;
  compressedGamestate.normsize = size;
  compressedGamestate.id = GAMESTATE;
  compressedGamestate.data = dest;
  
  return compressedGamestate;
}

GameState GameStateManager::decompress(GameStateCompressed a){
  int normsize = a.normsize;
  int compsize = a.compsize;
  unsigned char* dest = (unsigned char*)malloc( normsize );
  int retval;
  retval = uncompress( dest, &normsize, a.data, compsize );
  
  switch ( retval ) {
  case Z_OK: cout << "successfully compressed" << endl; break;
  case Z_MEM_ERROR: cout << "not enough memory available" << endl; break;
  case Z_BUF_ERROR: cout << "not enough memory available in the buffer" << endl; break;
  case Z_DATA_ERROR: cout << "data corrupted" << endl; break;
  }
  
  GameState gamestate = new GameState;
  gamestate.id = a.id;
  gamestate.size = normsize;
  gamestate.data = dest;
  
  return gamestate;
}

}


